// ***************************************************************************************
// File: listener.cpp
// Description: listener implementation
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;


NvListenerThread::NvListenerThread()
{
	b_Initialized = false;
	RequestToLeave = false;
	SessionId = "";
}



// Client thread main function
int NvListenerThread::Main()
{
	bool LocalRequestToLeave = false;
	RequestToLeave = false;
	int SleepTime = m_lListener->m_iSleepTime;
	string ProcName = m_lListener->ProcName;
	string ProcNameFailed = m_lListener->ProcNameFailed;
	while(1)
	{
		NV_LOCKOBJECT(this);
		LocalRequestToLeave = RequestToLeave;
		NV_UNLOCKOBJECT(this);
		if(LocalRequestToLeave)
			break;	// We must leave
		if(ProcName.length())
		{
			NvSession *Session = GetSession();
			if(Session != NULL)
			{
				// Do the main loop here
				// We prepare a command
				NvCommand Command;
				Command.m_iCommandSource = NV_SOURCE_SERVER;
				Command.m_sService = "SYSTEM";
				Command.m_sClass = "MISC";
				Command.m_sCommand = "NOP";
				Command.m_sInContainer = NV_CONTAINER_DEFAULT_NAME;
				Command.m_sOutContainer = NV_CONTAINER_DEFAULT_NAME;
				Command.Parameters.SetValue("NV_PROC", ProcName);


				// Run the procedure here
				bool Result = true;
				try
				{
					Result = Session->Command(&Command);
				}
				catch(...)
				{
					Result = false;
				}
				
				// Call the error procedure if one has been defined
				if(!Result && ProcNameFailed.length())
				{
					NvCommand ErrorCommand;
					ErrorCommand.m_iCommandSource = NV_SOURCE_SERVER;
					ErrorCommand.m_sService = "SYSTEM";
					ErrorCommand.m_sClass = "MISC";
					ErrorCommand.m_sCommand = "NOP";
					ErrorCommand.m_sInContainer = NV_CONTAINER_DEFAULT_NAME;
					ErrorCommand.m_sOutContainer = NV_CONTAINER_DEFAULT_NAME;
					ErrorCommand.Parameters.SetValue("NV_PROC", ProcNameFailed);
					char Buf[20];
					sprintf(Buf, "%d", Command.Error.m_iErrorCode);
					ErrorCommand.Parameters.SetValue("NV_ERROR_CODE", Buf);
					ErrorCommand.Parameters.SetValue("NV_ERROR_SERVICE", Command.Error.m_sErrorService.c_str());
					ErrorCommand.Parameters.SetValue("NV_ERROR_CLASS", Command.Error.m_sErrorClass.c_str());
					ErrorCommand.Parameters.SetValue("NV_ERROR_DESC", Command.Error.GetErrorDesc(m_lListener->Language.c_str()).c_str());
					ErrorCommand.Parameters.SetValue("NV_ERROR_INFO", Command.Error.m_sErrorInfo.c_str());
					try
					{
						Session->Command(&ErrorCommand);
					}
					catch(...)
					{
					}
				}

				// And free the session
				GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
			}
		}

		if(SleepTime > 0)
		{
			int LastSleepTime = 0;
			int NumLoops = 0;
			NumLoops = SleepTime / 1000;
			LastSleepTime = SleepTime % 1000;
			int i;
			for(i=0;i<NumLoops;i++)
			{
				SLEEP(1000);
				NV_LOCKOBJECT(this);
				LocalRequestToLeave = RequestToLeave;
				NV_UNLOCKOBJECT(this);
				if(LocalRequestToLeave)
					break;	// We must leave
			}
			if(LocalRequestToLeave)
				break;	// We must leave
			if(LastSleepTime > 0)
				SLEEP(LastSleepTime);
		}
	}
	return 0;
}


// called when an exception occurs (see class NvTask)
void NvListenerThread::OnException(string sInfo)
{
//	throw this;
}



// called before the Main function (see class NvTask)
bool NvListenerThread::OnInit()
{
	NvTask::OnInit();	// call the base class

	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_LISTENER);
	
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type LISTENER", NV_LOG_INFO, 1, sThreadId, "system");

	// Initialize the random generator for this thread
#ifdef WIN32
	int ThreadHandle = GetCurrentThreadId();						// Thread handle for window
#else
	pthread_t ThreadHandle = pthread_self();					// Thread handle for unix
#endif

	unsigned int randinit = (unsigned int)ThreadHandle;
	randinit *= (unsigned int)time(NULL);
	srand(randinit);

	b_Initialized = true;
	
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type LISTENER", NV_LOG_INFO, 1, sThreadId, "system");

	return true;
}




// Opens the session if not yet done
// If ever opened, return it
NvSession* NvListenerThread::GetSession()
{
	NvSession *Session = NULL;
	if(SessionId.length() != 0)
	{
		Session = GetMain()->Sessions.GetSession(SessionId.c_str());
		if(Session != NULL)
			return Session;
	}

	SessionId = "";

	// Create the session here
	if(GetMain()->CheckOverTime())
		return NULL;
	Session = GetMain()->Sessions.GetSession(NULL);
	if(Session == NULL)
		return NULL;

	string AppliNameHere = m_lListener->ListenerList->Application->GetName();
	CleanString(AppliNameHere);
	if(!AppliNameHere.length())
		AppliNameHere = "NVDEF";
	string UserHere = m_lListener->UserName;
	CleanString(UserHere);
	if(!UserHere.length())
		UserHere = "nvdef";
	string Hash = AppliNameHere + UserHere;
	Hash = MD5_string(Hash);
	MakeUpper(Hash);

	int iSessionTimeOut = -1;

	string OpenProc = m_lListener->OpenProc;
	CleanString(OpenProc);
	Session->m_sOpenProc = OpenProc;

	string CloseProc = m_lListener->CloseProc;
	CleanString(CloseProc);
	Session->m_sCloseProc = CloseProc;
	Session->m_bFromTransaction = true;

	string PasswordHere ="";

	GetMain()->AttachRequestList.Get(Session, ATTACH_REQUEST_TYPE_LISTENER, AppliNameHere);
	int LoginRetVal = Session->Login(UserHere.c_str(), PasswordHere.c_str(), AppliNameHere.c_str(), SESSION_TYPE_LISTENER, NULL, NULL, false, false, iSessionTimeOut);

	if(LoginRetVal != 1)
	{
		// We free the session
		GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());

		// And remove it
		GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());

		return NULL;
	}
	string UserHereForCheck = UserHere;
	CleanString(UserHereForCheck, true);
	MakeLower(UserHereForCheck);
	if(Session != NULL)
		if(UserHereForCheck == "nvadmin")
			Session->m_bAdministrator = true; // Administrator user
	
	SessionId = Session->m_sIdent;

	return Session;
}


// called after the Main function (see class NvTask)
void NvListenerThread::OnCleanup()
{
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type LISTENER", NV_LOG_INFO, 1, sThreadId, "system");

	// We must free the session here

	if(b_Initialized)
	{
		NvSession *Session = NULL;
		if(SessionId.length() != 0)
			Session = GetMain()->Sessions.GetSession(SessionId.c_str());
		if(Session != NULL)
		{
			if(Session->ActualCommand != NULL)
			{
				// The session is in a command so it's probably a crash
				if(Session->ActualCommand->m_sService != "SYSTEM")
					GetMain()->Services.FreeService(Session->ActualCommand->m_sService.c_str());
				Session->ActualCommand = NULL;
				Session->OriginalCommand = NULL;
				Session->InContainer = NULL;
				Session->OutContainer = NULL;
				Session->SetActualCommand("");
				Session->SetOriginalCommand("");
				NV_UNLOCKOBJECT(Session);
			}
			GetMain()->Sessions.FreeSession(Session->m_sIdent.c_str());
			GetMain()->Sessions.RemoveSession(Session->m_sIdent.c_str());
			SessionId = "";
		}

		// Detach thread from java interface if not yet made
		if(!m_bSameThread)
			GetMain()->JavaInterface.DetachJavaThread();
	}

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type LISTENER", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}


// called to stop the thread
// must be called from another thread
void NvListenerThread::Stop()
{
	NV_LOCKOBJECT(this);
	// We must set a flag at session level for the
	// SESSION:CHECK_CLOSE_REQUEST command to work
	// from inside the thread procedure
	NvSession *Session = NULL;
	if(SessionId.length() != 0)
		Session = GetMain()->Sessions.GetSession(SessionId.c_str());
	if(Session != NULL)
	{
		Session->SetListenerCloseRequest();
		GetMain()->Sessions.FreeSession(SessionId.c_str());
	}
	RequestToLeave = true;
	NV_UNLOCKOBJECT(this);
}
	

NvListener::NvListener()
{
	m_iNumThreads = 1;
	Name = "";
	Ident = "";
	Description = "";
	ProcName = "";
	ProcNameFailed = "";
	UserName = "";
	OpenProc = "";
	CloseProc = "";
	m_iSleepTime = 100;
	m_bRunning = false;
	m_bEnable = false;
	ListenerList = NULL;
	Language = "";
}
	

NvListener::~NvListener()
{
	Stop(true);
}
	

bool NvListener::Start()
{
	NV_LOCKOBJECT(this);
	if(m_bRunning)
	{
		// Ever running
		NV_UNLOCKOBJECT(this);
		return true;
	}
	if(m_iNumThreads == 0)
	{
		// No thread to launch
		m_bRunning = true;
		NV_UNLOCKOBJECT(this);
		return true;
	}

	int i;
	bool bError = false;
	for(i=0; i<m_iNumThreads; i++)
	{
		NvListenerThread *ListenerThread = new NvListenerThread;
		if(ListenerThread == NULL)
		{
			bError = true;
			break;
		}
		ListenerThread->m_lListener = this;

		if(!ListenerThread->Run())
		{
			delete ListenerThread;
			bError = true;
			break;
		}
		m_vTheads.push_back(ListenerThread);
	}
	if(!bError)
		m_bRunning = true;
	if(bError && m_vTheads.size())
	{
		m_bRunning = true;
		NV_UNLOCKOBJECT(this);
		Stop(true);
		return false;
	}
	NV_UNLOCKOBJECT(this);
	if(bError)
		return false;
	return true;
}

	

bool NvListener::Stop(bool Force, int WaitTime)
{
	NV_LOCKOBJECT(this);
	if(!m_bRunning)
	{
		// not running
		NV_UNLOCKOBJECT(this);
		return true;
	}

	list<NvListenerThread *>::iterator pos;
	for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
	{
		NvListenerThread *ListenerThread = *pos;
		if(ListenerThread == NULL)
			continue;
		ListenerThread->Stop();
	}
	
	// wait for them to stop
	time_t initial_time = time(NULL);
	bool bAtLeastOneRunning = false;
	while(1)
	{
		SLEEP(200);
		bAtLeastOneRunning = false;
		for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
		{
			NvListenerThread *ListenerThread = *pos;
			if(ListenerThread == NULL)
				continue;
			if(ListenerThread->IsThreadRunning())
			{
				bAtLeastOneRunning = true;
				break;
			}
		}
		if(!bAtLeastOneRunning)
			break;
		if((int)(time(NULL) - initial_time) > WaitTime)
			break;
	}
	if(bAtLeastOneRunning && Force)
	{
		// Let's kill hardly
		for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
		{
			NvListenerThread *ListenerThread = *pos;
			if(ListenerThread == NULL)
				continue;
			if(ListenerThread->IsThreadRunning())
				ListenerThread->Kill();
		}
	
		// wait for them to stop
		time_t initial_time = time(NULL);
		bAtLeastOneRunning = false;
		while(1)
		{
			SLEEP(200);
			bAtLeastOneRunning = false;
			for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
			{
				NvListenerThread *ListenerThread = *pos;
				if(ListenerThread == NULL)
					continue;
				if(ListenerThread->IsThreadRunning())
				{
					bAtLeastOneRunning = true;
					break;
				}
			}
			if(!bAtLeastOneRunning)
				break;
			if((int)(time(NULL) - initial_time) > WaitTime)
				break;
		}
	}
	
	if(bAtLeastOneRunning)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	for(pos = m_vTheads.begin(); pos != m_vTheads.end(); ++pos)
	{
		NvListenerThread *ListenerThread = *pos;
		if(ListenerThread != NULL)
			delete ListenerThread;
	}

	m_vTheads.clear();
	m_bRunning = false;
	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvListener::IsRunning()
{
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	RetVal = m_bRunning;
	NV_UNLOCKOBJECT(this);
	return RetVal;
}


// Constructor
NvListenerList::NvListenerList()
{
}

// Destructor
NvListenerList::~NvListenerList()
{
	Clear();
}



void NvListenerList::CleanListenerName(string& ListenerName)
{
	CleanString(ListenerName);

	string::size_type index;

	// remove stars
	while((index = ListenerName.find('*')) != string::npos)
		ListenerName.erase(index, 1);

	// remove slashes
	while((index = ListenerName.find('/')) != string::npos)
		ListenerName.erase(index, 1);

	// remove backslashes
	while((index = ListenerName.find('\\')) != string::npos)
		ListenerName.erase(index, 1);
}



string NvListenerList::GetListenerIdent(string ListenerName)
{
	CleanListenerName(ListenerName);
	MakeUpper(ListenerName);
	string ListenerIdent = MD5_string(ListenerName);
	MakeUpper(ListenerIdent);
	return ListenerIdent;
}



// Removes all listeners from the list
void NvListenerList::Clear()
{
	NV_LOCKOBJECT(this);
	map<string,NvListener *>::iterator pos;
	for(pos = m_mListeners.begin(); pos != m_mListeners.end(); ++pos)
	{
		NvListener *Listener = pos->second;
		if(Listener != NULL)
			delete Listener;
	}
	m_mListeners.clear();
	NV_UNLOCKOBJECT(this);
}


// used to get a listener (only internaly)
NvListener* NvListenerList::GetListener(const char *Name)
{
	NvListener *FoundListener = NULL;
	string NameHere = Name;
	string IdentHere = GetListenerIdent(NameHere);

	if(m_mListeners.find(IdentHere) != m_mListeners.end())
		FoundListener = m_mListeners[IdentHere];

	return FoundListener;
}


void NvListenerList::Load()
{
	NvRegistryContainer *Registry;
	Registry = &(Application->AppRegistry);

	// load the listener table
	// The listener table is stored in a table object named "listeners" in the
	// registry key system.listeners
	NvContainer Container;
	if(Registry->GetKey("system.listeners", &Container, NV_CONTAINER_COPY_NO_APPEND, "listeners", false, true))
	{
		NvObj *Object = Container.GetObject("listeners");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_TABLE))
		{
			NvObjTable *ListenerTable = (NvObjTable*)Object;
			int index_Name = ListenerTable->GetColumnIndex("Name");
			int index_Desc = ListenerTable->GetColumnIndex("Description");
			int index_ProcName = ListenerTable->GetColumnIndex("ProcName");
			int index_ProcNameFailed = ListenerTable->GetColumnIndex("ProcNameFailed");
			int index_UserName = ListenerTable->GetColumnIndex("UserName");
			int index_OpenProc = ListenerTable->GetColumnIndex("OpenProc");
			int index_CloseProc = ListenerTable->GetColumnIndex("CloseProc");
			int index_NumThreads = ListenerTable->GetColumnIndex("NumThreads");
			int index_SleepTime = ListenerTable->GetColumnIndex("SleepTime");
			int index_Enable = ListenerTable->GetColumnIndex("Enable");
			int index_Language = ListenerTable->GetColumnIndex("Language");
			int row_index;
			for(row_index=0; row_index<ListenerTable->GetNumRow(); row_index++)
			{
				// Get listener parameters
				string Name = ListenerTable->GetCellLine(row_index, index_Name);
				NvListenerList::CleanListenerName(Name);
				if(!Name.length()) // The listener is identified by its name
					continue;
				if(GetListener(Name.c_str()) != NULL)	// Check doubles
					continue;
				string Description = ListenerTable->GetCellLine(row_index, index_Desc);
				string ProcName = ListenerTable->GetCellLine(row_index, index_ProcName);
				string ProcNameFailed = ListenerTable->GetCellLine(row_index, index_ProcNameFailed);
				string UserName = ListenerTable->GetCellLine(row_index, index_UserName);
				string OpenProc = ListenerTable->GetCellLine(row_index, index_OpenProc);
				string CloseProc = ListenerTable->GetCellLine(row_index, index_CloseProc);
				string NumThreads = ListenerTable->GetCellLine(row_index, index_NumThreads);
				int iNumThreads = atoi(NumThreads.c_str());
				if(iNumThreads <= 0)
					iNumThreads = 1;
				string SleepTime = ListenerTable->GetCellLine(row_index, index_SleepTime);
				int iSleepTime = atoi(SleepTime.c_str());
				if(iSleepTime <= 0)
					iSleepTime = 0;
				string Enable = ListenerTable->GetCellLine(row_index, index_Enable);
				bool bEnable = false;
				if(Enable == "YES")
					bEnable = true;
				string Language = ListenerTable->GetCellLine(row_index, index_Language);

				CreateListener(Name.c_str(), Description.c_str(), ProcName.c_str(), ProcNameFailed.c_str(), UserName.c_str(), OpenProc.c_str(), CloseProc.c_str(),iNumThreads, iSleepTime, bEnable, Language.c_str());
			}
		}
		// Resave in case of change
		Save();
	}
}


// Saves the listener list to the system registry
// This function is called internally so its not locked
bool NvListenerList::Save()
{
	NvRegistryContainer *Registry;
	Registry = &(Application->AppRegistry);

	// First create the registry key "system.listeners" if it doesn't exist
	Registry->SetKey("system.listeners", NULL);

	NvContainer Container;
	if(!Registry->GetKey("system.listeners", &Container, NV_CONTAINER_COPY_NO_APPEND, NULL, false, true))
		return false; // cannot get the key. This should not occure here

	// Creates the listeners object if it doesn't exist
	Container.CreateObject("listeners", NV_OBJTYPE_TABLE);
	NvObj *Object = Container.GetObject("listeners");
	if(Object == NULL)
		return false; // This should not occur here

	if(Object->GetType() != NV_OBJTYPE_TABLE)
	{
		Container.RemoveObject("listeners");
		Container.CreateObject("listeners", NV_OBJTYPE_TABLE);
		Object = Container.GetObject("listeners");
		if(Object == NULL)
			return false; // This should not occur here
	}

	NvObjTable *ListenerTable = (NvObjTable*)Object;
	ListenerTable->Clear();
	ListenerTable->InsertColumn("Name");
	ListenerTable->InsertColumn("Description");
	ListenerTable->InsertColumn("ProcName");
	ListenerTable->InsertColumn("ProcNameFailed");
	ListenerTable->InsertColumn("UserName");
	ListenerTable->InsertColumn("OpenProc");
	ListenerTable->InsertColumn("CloseProc");
	ListenerTable->InsertColumn("NumThreads");
	ListenerTable->InsertColumn("SleepTime");
	ListenerTable->InsertColumn("Enable");
	ListenerTable->InsertColumn("Language");
	int index_Name = ListenerTable->GetColumnIndex("Name");
	int index_Desc = ListenerTable->GetColumnIndex("Description");
	int index_ProcName = ListenerTable->GetColumnIndex("ProcName");
	int index_ProcNameFailed = ListenerTable->GetColumnIndex("ProcNameFailed");
	int index_UserName = ListenerTable->GetColumnIndex("UserName");
	int index_OpenProc = ListenerTable->GetColumnIndex("OpenProc");
	int index_CloseProc = ListenerTable->GetColumnIndex("CloseProc");
	int index_NumThreads = ListenerTable->GetColumnIndex("NumThreads");
	int index_SleepTime = ListenerTable->GetColumnIndex("SleepTime");
	int index_Enable = ListenerTable->GetColumnIndex("Enable");
	int index_Language = ListenerTable->GetColumnIndex("Language");

	int j;
	map<string,NvListener *>::iterator pos;
	for(pos = m_mListeners.begin(),j=0; pos != m_mListeners.end(); ++pos)
	{
		NvListener *Listener = pos->second;
		if(Listener == NULL)
			continue;

		ListenerTable->InsertRow();
		ListenerTable->InsertCellLine(j, index_Name, Listener->Name.c_str());
		ListenerTable->InsertCellLine(j, index_Desc, Listener->Description.c_str());
		ListenerTable->InsertCellLine(j, index_ProcName, Listener->ProcName.c_str());
		ListenerTable->InsertCellLine(j, index_ProcNameFailed, Listener->ProcNameFailed.c_str());
		ListenerTable->InsertCellLine(j, index_UserName, Listener->UserName.c_str());
		ListenerTable->InsertCellLine(j, index_OpenProc, Listener->OpenProc.c_str());
		ListenerTable->InsertCellLine(j, index_CloseProc, Listener->CloseProc.c_str());
		char Buf[50];
		sprintf(Buf, "%d", Listener->m_iNumThreads);
		ListenerTable->InsertCellLine(j, index_NumThreads, Buf);
		sprintf(Buf, "%d", Listener->m_iSleepTime);
		ListenerTable->InsertCellLine(j, index_SleepTime, Buf);
		if(Listener->m_bEnable)
			ListenerTable->InsertCellLine(j, index_Enable, "YES");
		else
			ListenerTable->InsertCellLine(j, index_Enable, "NO");
		ListenerTable->InsertCellLine(j, index_Language, Listener->Language.c_str());
		j++;
	}
	
	if(!Registry->SetKey("system.listeners", &Container))
		return false;

	return true;
}
	


bool NvListenerList::CreateListener(const char *Name, const char *Description, const char *ProcName, const char *ProcNameFailed, const char *UserName, const char *OpenProc, const char *CloseProc, int NumThreads, int SleepTime, bool bEnable, const char *Language)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener != NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	FoundListener = new NvListener;
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	FoundListener->ListenerList = this;
	FoundListener->Name = NameHere;
	FoundListener->Ident = GetListenerIdent(NameHere);
	if(Description != NULL)
		FoundListener->Description = Description;
	if(ProcName != NULL)
		FoundListener->ProcName = ProcName;
	if(ProcNameFailed != NULL)
		FoundListener->ProcNameFailed = ProcNameFailed;
	if(UserName != NULL)
		FoundListener->UserName = UserName;
	if(OpenProc != NULL)
		FoundListener->OpenProc = OpenProc;
	if(CloseProc != NULL)
		FoundListener->CloseProc = CloseProc;
	FoundListener->m_iNumThreads = NumThreads;
	if(FoundListener->m_iNumThreads <= 0)
		FoundListener->m_iNumThreads = 0;
	FoundListener->m_iSleepTime = SleepTime;
	if(FoundListener->m_iSleepTime <= 0)
		FoundListener->m_iSleepTime = 0;
	FoundListener->m_bEnable = bEnable;
	if(Language != NULL)
		FoundListener->Language = Language;

	m_mListeners[FoundListener->Ident] = FoundListener;

	// Save the new listener list
	if(!Save())
	{
		m_mListeners.erase(FoundListener->Ident);
		delete FoundListener;
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKOBJECT(this);
	return true;
}


	


bool NvListenerList::SetListenerParam(const char *Name, const char *Description, const char *ProcName, const char *ProcNameFailed, const char *UserName, const char *OpenProc, const char *CloseProc, int NumThreads, int SleepTime)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(FoundListener->IsRunning())
	{
		// The listener must be stopped
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(Description != NULL)
		FoundListener->Description = Description;
	if(ProcName != NULL)
		FoundListener->ProcName = ProcName;
	if(ProcNameFailed != NULL)
		FoundListener->ProcNameFailed = ProcNameFailed;
	if(UserName != NULL)
		FoundListener->UserName = UserName;
	if(OpenProc != NULL)
		FoundListener->OpenProc = OpenProc;
	if(CloseProc != NULL)
		FoundListener->CloseProc = CloseProc;
	if(NumThreads != -1)
		FoundListener->m_iNumThreads = NumThreads;
	if(FoundListener->m_iNumThreads <= 0)
		FoundListener->m_iNumThreads = 1;
	if(SleepTime != -1)
		FoundListener->m_iSleepTime = SleepTime;
	if(FoundListener->m_iSleepTime <= 0)
		FoundListener->m_iSleepTime = 0;

	// Save the new listener list
	if(!Save())
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKOBJECT(this);
	return true;
}




bool NvListenerList::RemoveListener(const char *Name)
{
	if(Name ==NULL)
		return true;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return true;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}
	if(FoundListener->IsRunning())
	{
		// The listener must be stopped
		NV_UNLOCKOBJECT(this);
		return false;
	}
	m_mListeners.erase(FoundListener->Ident);
	delete FoundListener;


	// Save the new listener list
	Save();

	NV_UNLOCKOBJECT(this);
	return true;
}



bool NvListenerList::StartListener(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(FoundListener->IsRunning())
	{
		// The listener is ever running
		NV_UNLOCKOBJECT(this);
		return true;
	}

	bool RetVal = FoundListener->Start();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvListenerList::StopListener(const char *Name, bool Force, int WaitTime)
{
	if(Name ==NULL)
		return false;

	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(!FoundListener->IsRunning())
	{
		// The listener is ever stopped
		NV_UNLOCKOBJECT(this);
		return true;
	}

	bool RetVal = FoundListener->Stop(Force, WaitTime);

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvListenerList::EnableListener(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	FoundListener->m_bEnable = true;
	Save();

	NV_UNLOCKOBJECT(this);
	return true;
}




bool NvListenerList::DisableListener(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	FoundListener->m_bEnable = false;
	Save();

	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvListenerList::IsListenerRunning(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	bool RetVal = FoundListener->IsRunning();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}





bool NvListenerList::ListenerList(const char *Name, NvObjTable *ListenerInfo)
{
	ListenerInfo->Clear();

	bool AllListeners = true;

	NV_LOCKOBJECT(this);

	NvListener *FoundListener = NULL;
	if((Name != NULL) && strlen(Name))
	{
		// There is a request on a specific listener
		FoundListener = GetListener(Name);
		if(FoundListener == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllListeners = false;
	}


	ListenerInfo->InsertColumn("NAME");
	ListenerInfo->InsertColumn("DESCRIPTION");
	ListenerInfo->InsertColumn("PROC");
	ListenerInfo->InsertColumn("USER");
	ListenerInfo->InsertColumn("OPEN");
	ListenerInfo->InsertColumn("CLOSE");
	ListenerInfo->InsertColumn("THREADS");
	ListenerInfo->InsertColumn("SLEEP");
	ListenerInfo->InsertColumn("STATUS");
	ListenerInfo->InsertColumn("PROCF");

	if(AllListeners)
	{
		int j;
		map<string,NvListener *>::iterator pos;
		for(pos = m_mListeners.begin(),j=0; pos != m_mListeners.end(); ++pos)
		{
			NvListener *Listener = pos->second;
			if(Listener == NULL)
				continue;
			ListenerInfo->InsertRow();
			ListenerInfo->InsertCellLine(j, 0, Listener->Name);
			ListenerInfo->InsertCellLine(j, 1, Listener->Description);
			ListenerInfo->InsertCellLine(j, 2, Listener->ProcName);
			ListenerInfo->InsertCellLine(j, 3, Listener->UserName);
			ListenerInfo->InsertCellLine(j, 4, Listener->OpenProc);
			ListenerInfo->InsertCellLine(j, 5, Listener->CloseProc);
			char Buffer[20];
			sprintf(Buffer, "%d", Listener->m_iNumThreads);
			ListenerInfo->InsertCellLine(j, 6, Buffer);
			sprintf(Buffer, "%d", Listener->m_iSleepTime);
			ListenerInfo->InsertCellLine(j, 7, Buffer);
			if(Listener->IsRunning())
				ListenerInfo->InsertCellLine(j, 8, "RUNNING");
			else
				ListenerInfo->InsertCellLine(j, 8, "STOPPED");
			ListenerInfo->InsertCellLine(j, 9, Listener->ProcNameFailed);
			j++;
		}
	}
	else
	{
		ListenerInfo->InsertRow();
		ListenerInfo->InsertCellLine(0, 0, FoundListener->Name);
		ListenerInfo->InsertCellLine(0, 1, FoundListener->Description);
		ListenerInfo->InsertCellLine(0, 2, FoundListener->ProcName);
		ListenerInfo->InsertCellLine(0, 3, FoundListener->UserName);
		ListenerInfo->InsertCellLine(0, 4, FoundListener->OpenProc);
		ListenerInfo->InsertCellLine(0, 5, FoundListener->CloseProc);
		char Buffer[20];
		sprintf(Buffer, "%d", FoundListener->m_iNumThreads);
		ListenerInfo->InsertCellLine(0, 6, Buffer);
		sprintf(Buffer, "%d", FoundListener->m_iSleepTime);
		ListenerInfo->InsertCellLine(0, 7, Buffer);
		if(FoundListener->IsRunning())
			ListenerInfo->InsertCellLine(0, 8, "RUNNING");
		else
			ListenerInfo->InsertCellLine(0, 8, "STOPPED");
		ListenerInfo->InsertCellLine(0, 9, FoundListener->ProcNameFailed);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}




bool NvListenerList::StartListeners()
{
	NV_LOCKOBJECT(this);
	map<string,NvListener *>::iterator pos;
	for(pos = m_mListeners.begin(); pos != m_mListeners.end(); ++pos)
	{
		NvListener *Listener = pos->second;
		if(Listener == NULL)
			continue;
		if(Listener->m_bEnable)
			Listener->Start();
	}
	NV_UNLOCKOBJECT(this);

	return true;
}




bool NvListenerList::StopListeners()
{

	NV_LOCKOBJECT(this);
	map<string,NvListener *>::iterator pos;
	for(pos = m_mListeners.begin(); pos != m_mListeners.end(); ++pos)
	{
		NvListener *Listener = pos->second;
		if(Listener == NULL)
			continue;
		Listener->Stop(true, 30);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvListenerList::ListenerExist(const char *Name)
{
	if(Name ==NULL)
		return false;
	string NameHere = Name;
	CleanListenerName(NameHere);

	if(!NameHere.length())
		return false;

	bool RetVal = false;
	NV_LOCKOBJECT(this);
	NvListener *FoundListener = GetListener(NameHere.c_str());
	if(FoundListener != NULL)
		RetVal = true;

	NV_UNLOCKOBJECT(this);
	return RetVal;
}
